// Copyright (c) Microsoft Corporation.  All rights reserved.
//
using System;
using System.Configuration;
using System.Collections.Generic;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Xsl;
using System.Text.RegularExpressions;

namespace Microsoft.Ddue.Tools
{
    /// <summary>
    /// Provides support for snippets
    /// </summary>
    public class ExampleResolverComponent : BuildComponent
    {
		private XPathExpression codeLinkQuery;

		private XmlNamespaceManager nsManager = new XmlNamespaceManager(new NameTable());

		public ExampleResolverComponent(XPathNavigator configuration)
            : base(configuration)
        {
            XPathNodeIterator contentNodes = configuration.Select("content");
            foreach (XPathNavigator contentNode in contentNodes)
            {
                string file = contentNode.GetAttribute("file", String.Empty);
                if (string.IsNullOrEmpty(file))
                    throw new ConfigurationErrorsException("When instantiating an example component, you must use the file attribute to specify a shared content file.");
				LoadContent(file);
            }
			WriteMessage(MessageLevel.Info, String.Format("Loaded {0} snippets", snippets.Count));

			// set up the context
			// put in a default entry for ddue
			nsManager.AddNamespace("ddue", "http://ddue.schemas.microsoft.com/authoring/2003/6");
			// look for context nodes, which could override the default
			XPathNodeIterator contextNodes = configuration.Select("context");
			foreach (XPathNavigator contextNode in contextNodes)
			{
				string prefix = contextNode.GetAttribute("prefix", String.Empty);
				string uri = contextNode.GetAttribute("name", String.Empty);
				// add the namespace's prefix and uri. If the prefix exists in the manager, this replaces it.
				nsManager.AddNamespace(prefix, uri);
			}

			// initialize the xpath expression
			string xpathString = "//ddue:codeReference";
			// look for xpath nodes, which could override the default
			XPathNodeIterator xpathNodes = configuration.Select("xpath");
			foreach (XPathNavigator xpathNode in xpathNodes)
			{
				string expression = xpathNode.GetAttribute("expression", String.Empty);
				if (expression != string.Empty)
					xpathString = expression;
			}
			codeLinkQuery = XPathExpression.Compile(xpathString);

			codeLinkQuery.SetContext(nsManager);
		}

		/// <summary>
		/// Loads the code example shared content file. 
		/// </summary>
		/// <param name="file">The shared content file to load.</param>
		private void LoadContent(string file)
		{
			WriteMessage(MessageLevel.Info, String.Format("Loading snippet file '{0}'.", file));
			XmlReader reader = XmlReader.Create(file);
			try
			{
				reader.MoveToContent();
				while (!reader.EOF)
				{

					if ((reader.NodeType == XmlNodeType.Element) && (reader.Name == "item"))
					{
						string key = reader.GetAttribute("id");
						if (key != string.Empty)
						{
							string value = reader.ReadInnerXml();

							// use lower case, case-insensitive IDs to reduce typo problems for authors
							// for duplicate item ids, the last one wins
							if (snippets.ContainsKey(key.ToLower()))
							{
								snippets.Remove(key.ToLower());
								WriteMessage(MessageLevel.Warn, string.Format("Duplicate snippet item '{0}'. Last one wins!", key));
							}
							// add the item
							snippets.Add(key.ToLower(), value);
						}
					}
					else
					{
						reader.Read();
					}
				}
			}
			finally
			{
				reader.Close();
			}
		}

		private string currentKey = string.Empty;

		public override void Apply(XmlDocument document, string key)
        {
			currentKey = key;

			XPathNodeIterator referenceNodes = document.CreateNavigator().Select(codeLinkQuery);
            foreach (XPathNavigator referenceNode in referenceNodes)
            {
                string referenceText = referenceNode.Value;
				if (referenceText == string.Empty)
				{
					WriteMessage(MessageLevel.Warn, string.Format("Invalid codeReference: missing snippet ID. Topic id: '{0}'", currentKey));
					continue;
				}

				string[] poundSignSplit = referenceText.Split('#');
				if (poundSignSplit.Length != 2 || poundSignSplit[0] == "" || poundSignSplit[1] == "")
				{
					WriteMessage(MessageLevel.Warn, string.Format("Invalid snippet reference: {0}. Topic id: '{1}'", referenceText, currentKey));
					continue;
				}

				string[] snippetSuffixes = poundSignSplit[1].Split(',');

				if (snippetSuffixes.Length > 1)
				{
					ProcessCombinedSnippet(document, referenceNode, poundSignSplit[0], snippetSuffixes);
				}
				else
				{
					ProcessSnippetID(document, referenceNode, referenceText);
				}
            }
        }

		private void ProcessSnippetID(XmlDocument xdoc, XPathNavigator codeReference, string snippetId)
		{
			string snippetContent;
			if (snippets.TryGetValue(snippetId.ToLower(), out snippetContent))
				AppendSampleCode(xdoc, codeReference, snippetContent);
			else
			{
				WriteMessage(MessageLevel.Warn, string.Format("Unresolved snippet reference: {0}. Topic id: '{1}'", snippetId, currentKey));
				codeReference.InnerXml = "";
			}
		}

		private void ProcessCombinedSnippet(XmlDocument xdoc, XPathNavigator codeReference, string idPrefix, string[] idSuffixes)
		{
			Dictionary<string, XmlDocumentFragment> samplesCollection = new Dictionary<string, XmlDocumentFragment>();
			List<XPathNavigator> noDevlangSnippets = new List<XPathNavigator>();

			XmlDocumentFragment docFragment;

			bool snippetFound = false;

			foreach (string snippetSuffix in idSuffixes)
			{
				string snippetId = idPrefix + "#" + snippetSuffix;
				string snippetContent;
				if (snippets.TryGetValue(snippetId.ToLower(), out snippetContent))
				{
					snippetFound = true;
					docFragment = xdoc.CreateDocumentFragment();
					docFragment.InnerXml = snippetContent;
					XPathNodeIterator samples = docFragment.CreateNavigator().Select("//sampleCode");
					foreach (XPathNavigator sample in samples)
					{
						string devlang = sample.GetAttribute("language", string.Empty);

						if (devlang == string.Empty)
						{
							noDevlangSnippets.Add(sample);
						}
						else
						{
							XmlDocumentFragment combinedSample;
							if (samplesCollection.TryGetValue(devlang, out combinedSample))
							{
								combinedSample.InnerXml += "<include item=\"snippetDivider\"/>" + sample.InnerXml;
								samplesCollection.Remove(devlang);
								samplesCollection.Add(devlang, combinedSample);
							}
							else
							{
								combinedSample = xdoc.CreateDocumentFragment();
								samplesCollection.Add(devlang, combinedSample);
							}
						}
					}
				}
				else
				{
					WriteMessage(MessageLevel.Warn, string.Format("Unresolved snippet reference: {0}. Topic id: '{1}'", snippetId, currentKey));
				}
			}

			if (snippetFound)
			{
				// Combine samples by language.
				foreach (XmlDocumentFragment sampleFrag in samplesCollection.Values)
				{
					AppendSampleCode(codeReference, sampleFrag);
				}

				// Handle samples not marked with a language.
				foreach (XPathNavigator sample in noDevlangSnippets)
					AppendSampleCode(xdoc, codeReference, sample.OuterXml);
			}
			else
			{
				codeReference.InnerXml = "";
			}
		}

		private void AppendSampleCode(XmlDocument xdoc, XPathNavigator codeReference, string newText)
		{
			XmlDocumentFragment sampleFrag = xdoc.CreateDocumentFragment();
			sampleFrag.InnerXml = newText;
			AppendSampleCode(codeReference, sampleFrag);
		}

		private void AppendSampleCode(XPathNavigator codeReference, XmlDocumentFragment sampleFrag)
		{
			RemoveLeadingWhiteSpace(sampleFrag);
			codeReference.InnerXml = sampleFrag.OuterXml;
		}

		private void RemoveLeadingWhiteSpace(XmlDocumentFragment sampleXmlFragment)
		{
			//Get all of the <sampleCode> elements
			XmlNodeList samples = sampleXmlFragment.SelectNodes("//sampleCode");
			foreach (XmlNode sample in samples)
			{
				//Normalize the whitespace first. One Tab == Four Spaces
				string innerText = sample.InnerText.Replace("\t", "    ");

				//Look at each line and see which one has the least amount of whitespace in front of it.
				string[] lines = innerText.Split('\r', '\n');
				int iMinSpaces = Int32.MaxValue;
				foreach (string line in lines)
				{
					if (line.Trim().Length > 0)
					{
						int spaces = Regex.Match(line, @"^\s*", RegexOptions.None).Value.Length;
						iMinSpaces = (spaces < iMinSpaces ? spaces : iMinSpaces);
					}
				}

				if ((iMinSpaces != Int32.MaxValue) && (iMinSpaces > 0))
				{
					string regex = @"(^|\r\n|\n|\r) {" + iMinSpaces.ToString() + "}";
					sample.InnerText = Regex.Replace(innerText, regex, @"$1", RegexOptions.None);
				}
			}
		}

		/// <summary>
        /// The collection of snippets. 
        /// </summary>
        private Dictionary<string, string> snippets = new Dictionary<string, string>();
    }

}