// Copyright (c) Microsoft Corporation.  All rights reserved.
//
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Xml;
using System.Xml.XPath;


namespace Microsoft.Ddue.Tools
{

	public class MrefLinkResolverComponent : BuildComponent
	{

		// Instantiation logic

		public MrefLinkResolverComponent(XPathNavigator configuration)
			: base(configuration)
		{
			XPathNodeIterator targets_nodes = configuration.Select("targets");
			foreach (XPathNavigator targets_node in targets_nodes)
			{
				string filespec = targets_node.GetAttribute("file", String.Empty);
				string type_value = targets_node.GetAttribute("type", String.Empty);
				string topicXPath = targets_node.GetAttribute("topicXPath", String.Empty);

				ReferenceLinkType type = (ReferenceLinkType)Enum.Parse(typeof(ReferenceLinkType), type_value, true);

				string[] files = null;
				if (File.Exists(filespec))
				{
					// we're loading a single file
					AddTargets(filespec, type, topicXPath);
				}
				else
				{
					// must be loading a set of files
					if (Directory.Exists(filespec))
					{
						// if they specified a directory, transform all the files in the directory
						files = Directory.GetFiles(filespec);
					}
					else
					{
						// it's not a file or a directory, maybe it's a path with wildcards
						string directoryPath = Path.GetDirectoryName(filespec);
						string filePath = Path.GetFileName(filespec);
						files = Directory.GetFiles(directoryPath, filePath);
					}
					foreach (string file in files)
					{
						// load targets from each file
						AddTargets(file, type, topicXPath);
					}
				}
				WriteMessage(MessageLevel.Info, String.Format("Indexed {0} reference link targets.", index.Count));
			}
		}

		private void AddTargets(string file, ReferenceLinkType type, string topicXPath)
		{
			topicXPath = (topicXPath == string.Empty) ? "/topics/topic" : topicXPath;

			XPathDocument document = new XPathDocument(file);
			XPathNodeIterator target_nodes = document.CreateNavigator().Select(topicXPath);

			foreach (XPathNavigator targetNode in target_nodes)
			{
				string id = targetNode.GetAttribute("id", String.Empty);
				if (String.IsNullOrEmpty(id)) continue;

				ReferenceLink link;

				switch (type)
				{
					case ReferenceLinkType.Local:
						// is this id already in the Dictionary?
						if (index.TryGetValue(id, out link))
						{
							// An entry with this id is already in the dictionary. Is it local or index or none?
							if ((link as ReferenceLinkIndex) != null || (link as ReferenceLinkNone) != null)
							{
								// a local target trumps a non-local one; remove the non-local, add the local
								index.Remove(id);
								index.Add(id, new ReferenceLinkLocal(targetNode));
							}
							else
							{
								// already have a local link with this id
								WriteMessage(MessageLevel.Warn, string.Format("Duplicate topic id '{0}'.", id));
							}
						}
						else
							index.Add(id, new ReferenceLinkLocal(targetNode));
						break;

					case ReferenceLinkType.Index:
						// is this id already in the Dictionary?
						if (index.TryGetValue(id, out link))
						{
							// if this is duplicate of a local link, don't worry about it
							// otherwise, do a warning for duplicate index links
							if ((link as ReferenceLinkLocal) == null)
								WriteMessage(MessageLevel.Warn, string.Format("Duplicate topic id '{0}'.", id));
						}
						else
							index.Add(id, new ReferenceLinkIndex(targetNode));
						break;

					case ReferenceLinkType.None:
						// these are links that we intend to be cold, e.g. for external links in a standalone hxs
						// is this id already in the Dictionary?
						if (index.TryGetValue(id, out link))
						{
							// if this is duplicate of a local link, don't worry about it
							// otherwise, do a warning for duplicate index links
							if ((link as ReferenceLinkLocal) == null)
								WriteMessage(MessageLevel.Warn, string.Format("Duplicate topic id '{0}'.", id));
						}
						else
							index.Add(id, new ReferenceLinkNone(targetNode));
						break;
				}
			}
		}

		private Dictionary<string, ReferenceLink> index = new Dictionary<string, ReferenceLink>();

		// The action of the component
		private string currentKey = string.Empty;
		public override void Apply(XmlDocument document, string key)
		{
			currentKey = key;
			XmlNodeList linkNodes = document.SelectNodes("//referenceLink");

			List<string> precedingHotLinks = new List<string>();
			foreach (XmlElement linkNode in linkNodes)
			{
				string target = linkNode.GetAttribute("target");
				if (String.IsNullOrEmpty(target)) continue;

				string forceHotVal = linkNode.GetAttribute("forceHot");
				bool forceHot = (string.IsNullOrEmpty(forceHotVal)) ? false : Convert.ToBoolean(forceHotVal);

				string qualifiedVal = linkNode.GetAttribute("qualified");
				bool qualified = (string.IsNullOrEmpty(qualifiedVal)) ? false : Convert.ToBoolean(qualifiedVal);

				string showParametersVal = linkNode.GetAttribute("showParameters");
				bool showParameters = (string.IsNullOrEmpty(showParametersVal)) ? false : Convert.ToBoolean(showParametersVal);

				string autoUpgradeVal = linkNode.GetAttribute("autoUpgrade");
				bool autoUpgrade = (string.IsNullOrEmpty(autoUpgradeVal)) ? true : Convert.ToBoolean(autoUpgradeVal);

				string title = linkNode.InnerText.Trim();

				bool isHot = (target != key && (forceHot || !precedingHotLinks.Contains(target)));

				XmlElement linkElement;
				ReferenceLink link;
				if (index.TryGetValue(target, out link))
				{
					title = (String.IsNullOrEmpty(title)) ? link.GetLinkText(qualified, autoUpgrade, showParameters) : title;
					if (isHot)
						linkElement = link.GetLinkElement(document, title, autoUpgrade);
					else
						linkElement = ColdLink(document, title);
				}
				else
				{
					title = String.IsNullOrEmpty(title) ? GetReferenceTitle(target, qualified, autoUpgrade) : title;
					linkElement = ColdLink(document, title);
					isHot = false;
					WriteMessage(MessageLevel.Warn, string.Format("Id '{0}'. Display text '{1}'. Topic id: '{2}'", target, title, currentKey));
					//WriteMessage(MessageLevel.Warn, "unresolvedMrefLink", string.Format("Id '{0}'. Display text '{1}'. Topic id: '{2}'", target, title, currentKey));
				}
				linkNode.ParentNode.ReplaceChild(linkElement, linkNode);
				if (isHot)
					precedingHotLinks.Add(target);
			}
		}

		private XmlElement ColdLink(XmlDocument document, string displayText)
		{
			XmlElement linkElement = document.CreateElement("span");
			XmlText link_text = document.CreateTextNode(displayText);
			linkElement.SetAttribute("class", "referenceNoLink");
			linkElement.AppendChild(link_text);
			return linkElement;
		}

		private XPathExpression referenceLinks = XPathExpression.Compile("//referenceLink");

		private string GetReferenceTitle(string target, bool qualified, bool autoUpgrade)
		{
			string[] idSplit = target.Split(':');
			if (idSplit.Length != 2 || idSplit[0] == "" || idSplit[1] == "")
				return (target);
			string prefix = idSplit[0];
			string fullName = idSplit[1];
			switch (prefix)
			{
				case "AllMembers.T":
					return GetTypeShortName(fullName) + "Members";
				case "Methods.T":
					return GetTypeShortName(fullName) + "Methods";
				case "Fields.T":
					return GetTypeShortName(fullName) + "Fields";
				case "Properties.T":
					return GetTypeShortName(fullName) + "Properties";
				case "Events.T":
					return GetTypeShortName(fullName) + "Events";
				case "Operators.T":
					return GetTypeShortName(fullName) + "Operators";
				case "AttachedProperties.T":
					return GetTypeShortName(fullName) + "Attached Properties";
				case "AttachedEvents.T":
					return GetTypeShortName(fullName) + "Attached Events";
				case "DerivedTypeList.T":
					return GetTypeShortName(fullName) + "Derived Classes";
				case "N":
					return fullName;
				case "T":
					return (GetTypeTitle(fullName, qualified));
				case "M":
					return (GetMethodTitle(fullName, qualified));
				case "P":
					return (GetPropertyTitle(fullName, qualified));
				case "F":
				case "E":
					return (GetSimpleMemberTitle(fullName, qualified));
			}
			return (target);
		}

		private string GetTypeShortName(string typeFullName)
		{
			string[] dotSplit = typeFullName.Split(('.'));
			if (dotSplit.Length < 2 || dotSplit[dotSplit.Length - 1] == "")
				return typeFullName;

			string shortName = dotSplit[dotSplit.Length - 1];
			string[] tickSplit = shortName.Split('`');
			return tickSplit[0];
		}

		private string GetTypeTitle (string reference, bool qualified) {
			// Console.WriteLine(reference);
			int length = reference.Length;
			if (length == 0) return(reference);
			
			if (reference.EndsWith("[]")) {
				return( GetTypeTitle(reference.Substring(0,length-2), qualified) + "[]" );
			}

			// strip reference indicators
			if (reference.EndsWith("@")) {
				return( GetTypeTitle(reference.Substring(0,length-1), qualified) );
			}

			if (reference[0] == '`') {
				// a pure generic template parameter
				if ((length > 1) && (reference[1] == '`')) {
					string parameter = reference.Substring(2);
					return( String.Format("T", parameter) );
				} else {
					string parameter = reference.Substring(1);
					return( String.Format("T", parameter) );
				}
			} else if (reference[length-1] == '}') {
				// a specialized generic
				int index = reference.LastIndexOf('{');
				if (index < 0) return(String.Empty);
				string genericType = reference.Substring(0,index);
				string specialization = reference.Substring(index+1,length-index-2);
				return( String.Format("{0}<{1}>", GetTypeTitle(genericType, qualified), CombineTypes(SeperateTypes(specialization))) );
			} else {
				// other types
				if ((length > 2) && (reference[length-2] == '`')) {
					// an unspecialized generic type
					string parameterCount = reference.Substring(length-1);
					string baseType = reference.Substring(0,length-3);
					return( GetTypeTitle(reference.Substring(0,length-2), qualified) + GetTemplateParameters("T:" + reference) );
				} else {
					// a simple, straightforward type
					if (qualified) {
						return(reference);
					} else {
						int index = reference.LastIndexOf('.');
						if (index < 0) {
							return(reference);
						} else {
							string type = reference.Substring(index+1);
							return(type);
						}
					}
				}
			}
		}

		private string GetTemplateParameters (string api) {
			if (api.EndsWith("`1")) {
				return("<T>");
			} else if (api.EndsWith("`2")) {
				return("<T, U>");
			} else if (api.EndsWith("`3")) {
				return("<T, U, V>");
			} else {
				return("<UNK>");
			}
		}

		private string GetSimpleMemberTitle (string reference, bool qualified) {
			int index = reference.LastIndexOf('.');
			if (index < 0) return(reference);
			string member = reference.Substring(index+1);
			int length = member.Length;
			if ((length > 3) && (member[length-3] == '`') && (member[length-2] == '`')) {
				// for generic methods, replace tic marks
				member = member.Substring(0,length-3) + GetTemplateParameters(member.Substring(length-3));
			} else if (member == "#ctor") {
				// for constructors, replace #ctor with type name
				member = GetTypeTitle(reference.Substring(0,index), false);
			}

			if (qualified) {
				string type = GetTypeTitle(reference.Substring(0,index), false);
				return( String.Format("{0}.{1}", type, member) );
			} else {
				return(member);
			}
		}

		private string GetParameterizedMemberTitle (string reference) {
			int index = reference.IndexOf('(');
			if (index < 0) return(reference);
			string member_reference = reference.Substring(0, index);
			string parameters = reference.Substring(index+1, reference.Length-index-2);

			string[] parameter_references = SeperateTypes(parameters);

			StringBuilder title = new StringBuilder( GetSimpleMemberTitle(member_reference, false) );
			title.Append('(');
			for(int i=0; i<parameter_references.Length; i++) {
				if (i>0) title.Append(", ");
				title.Append( GetTypeTitle(parameter_references[i], false) );
			}
			title.Append(')');
			return(title.ToString());			
		}

		private string GetPropertyTitle (string reference, bool qualified) {
			if (reference.EndsWith(")")) {
				return( GetParameterizedMemberTitle(reference) );
			} else {
				return( GetSimpleMemberTitle(reference, qualified) );
			}
		}

		private string GetMethodTitle (string reference, bool qualified) {
			if (reference.EndsWith(")")) {
				return( GetParameterizedMemberTitle(reference) );
			} else {
				return( GetSimpleMemberTitle(reference, qualified) + "()" );
			}
		}

		private string[] SeperateTypes (string typelist) {
			List<string> types = new List<string>();

			int start = 0;
			int specializationCount = 0;
			for (int index=0; index<typelist.Length; index++) {
				switch (typelist[index]) {
					case '{':
						specializationCount++;
					break;
					case '}':
						specializationCount--;
					break;
					case ',':
						if (specializationCount == 0) {
							types.Add( typelist.Substring(start, index-start) );
							start = index + 1;
						}
					break;
				}
			}
			types.Add( typelist.Substring(start) );
			return(types.ToArray());
		}

		private string CombineTypes(string[] types)
		{
			StringBuilder text = new StringBuilder();
			for (int i = 0; i < types.Length; i++)
			{
				if (i > 0) text.Append(", ");
				text.Append(GetTypeTitle(types[i], false));
			}
			return (text.ToString());
		}
	}

	internal abstract class ReferenceLink
	{
		/*
    <topic id="{@id}" filename="{file/@name}">
      <xsl:copy-of select="apidata/@name"/>
      <xsl:copy-of select="apidata/@group"/>
      <xsl:copy-of select="apidata/@subgroup"/>
      <xsl:copy-of select="apidata/@subsubgroup"/>
      <xsl:if test="$namespace!=''">
        <xsl:attribute name="namespace">
          <xsl:value-of select="$namespace"/>
        </xsl:attribute>
      </xsl:if>
      <xsl:if test="$declaringType!=''">
        <xsl:attribute name="declaringType">
          <xsl:value-of select="$declaringType"/>
        </xsl:attribute>
      </xsl:if>
      <xsl:call-template name="parameters"/>
      <xsl:copy-of select="templates"/>
    </topic>
		 */

		protected string targetId = string.Empty;
		protected string shortName = string.Empty;
		protected string qualifiedName = string.Empty;
		protected string group = string.Empty;
		protected string subgroup = string.Empty;
		protected string subsubgroup = string.Empty;
		protected string pseudo = string.Empty;
		protected string parameterList = string.Empty;
		protected string templateList = string.Empty;
		protected string overloadTopicId = string.Empty;

		public ReferenceLink(XPathNavigator targetNode)
		{
			string id = targetNode.GetAttribute("id", string.Empty);
			targetId = id;
			string namespaceName = targetNode.GetAttribute("namespace", string.Empty);
			string declaringType = targetNode.GetAttribute("declaringType", string.Empty);

			group = targetNode.GetAttribute("group", string.Empty);
			subgroup = targetNode.GetAttribute("subgroup", string.Empty);
			subsubgroup = targetNode.GetAttribute("subsubgroup", string.Empty);
			pseudo = targetNode.GetAttribute("pseudo", string.Empty);

			shortName = targetNode.GetAttribute("name", string.Empty);
			switch (group)
			{
				case "type":
					qualifiedName = namespaceName + "." + shortName;
					break;
				case "member":
					qualifiedName = declaringType + "." + shortName;
					break;
				case "namespace":
					qualifiedName = shortName;
					break;
			}

			// get information about the overload list topic, for use with autoUpgrade="true" links
			XPathNavigator overloadTopicNode = targetNode.SelectSingleNode("overloadTopic");
			if (overloadTopicNode != null)
				overloadTopicId = overloadTopicNode.GetAttribute("id", string.Empty);

			// build a parameter list string if there are params
			XPathNavigator parametersNode = targetNode.SelectSingleNode("parameters");
			if (parametersNode != null)
			{
				StringBuilder sb = new StringBuilder("(");
				XPathNodeIterator parameters = parametersNode.Select("parameter");
				int curr = 0;
				foreach (XPathNavigator parameter in parameters)
				{
					sb = sb.Append(parameter.GetAttribute("typeName", string.Empty));
					if (++curr < parameters.Count)
						sb = sb.Append(", ");
				}
				sb = sb.Append(")");
				parameterList = sb.ToString();
			}

			// build a type parameter list string if there are type params
			XPathNavigator typeParametersNode = targetNode.SelectSingleNode("templates");
			if (typeParametersNode != null)
			{
				StringBuilder sb = new StringBuilder("<");
				XPathNodeIterator typeParameters = typeParametersNode.Select("template");
				int curr = 0;
				foreach (XPathNavigator typeParameter in typeParameters)
				{
					sb = sb.Append(typeParameter.GetAttribute("name", string.Empty));
					if (++curr < typeParameters.Count)
						sb = sb.Append(", ");
				}
				sb = sb.Append(">");
				templateList = sb.ToString();
			}
		}

		public virtual string GetLinkText(bool qualified, bool autoUpgrade, bool showParameters)
		{
			StringBuilder sb = new StringBuilder(qualified ? qualifiedName + templateList : shortName + templateList);

			// for list topics include a descriptor
			// this is just a failsafe that shouldn't be used; XSLT should pass in a loc'd display text
			if (group == "type" && !string.IsNullOrEmpty(pseudo))
			{
				switch (subgroup)
				{
					case "AllMembers":
						sb = sb.Append(" Members");
						break;
					case "AttachedProperties":
						sb = sb.Append(" Attached Properties");
						break;
					case "AttachedEvents":
						sb = sb.Append(" Attached Events");
						break;
					default:
						sb = sb.Append(string.Format(" {0}", subgroup));
						break;
				}
			}

			// if showParameters is true, add the param list, if any
			if (showParameters)
				sb = sb.Append(parameterList);

			return sb.ToString();
		}

		public abstract XmlElement GetLinkElement(XmlDocument document, string text, bool autoUpgrade);
	}

	internal class ReferenceLinkLocal : ReferenceLink
	{

		public ReferenceLinkLocal(XPathNavigator targetNode)
			: base(targetNode)
		{
			this.targetFile = targetNode.GetAttribute("filename", string.Empty);
			
			// get information about the overload list topic, for use with autoUpgrade="true" links
			XPathNavigator overloadTopicNode = targetNode.SelectSingleNode("overloadTopic");
			if (overloadTopicNode != null)
			{
				overloadTopicFilename = overloadTopicNode.GetAttribute("filename", string.Empty);
			}
		}

		private string targetFile = string.Empty;
		protected string overloadTopicFilename = string.Empty;

		public override XmlElement GetLinkElement(XmlDocument document, string text, bool autoUpgrade)
		{
			string file = (autoUpgrade && !string.IsNullOrEmpty(overloadTopicFilename)) ? overloadTopicFilename : targetFile;
			XmlElement link = document.CreateElement("a");
			XmlText link_text = document.CreateTextNode(text);
			link.SetAttribute("href", file + ".htm");
			link.AppendChild(link_text);
			return (link);
		}

	}

	internal class ReferenceLinkIndex : ReferenceLink
	{

		public ReferenceLinkIndex(XPathNavigator targetNode) 
			: base(targetNode)
		{
		}

		private static string mshelp = "http://msdn.microsoft.com/mshelp";

		public override XmlElement GetLinkElement(XmlDocument document, string text, bool autoUpgrade)
		{
			// autoUpgrade=true means link to the overload topic; false means link to signature topic
			string id = (autoUpgrade && !string.IsNullOrEmpty(overloadTopicId)) ? overloadTopicId : targetId;
			XmlElement link = document.CreateElement("mshelp", "link", mshelp);
			XmlText link_text = document.CreateTextNode(text);
			link.SetAttribute("keywords", id);
			link.SetAttribute("tabindex", "0");
			link.AppendChild(link_text);
			return (link);
		}

	}

	internal class ReferenceLinkNone : ReferenceLink
	{

		public ReferenceLinkNone(XPathNavigator targetNode) 
			: base(targetNode)
		{
		}

		public override XmlElement GetLinkElement(XmlDocument document, string text, bool autoUpgrade)
		{
			XmlElement linkElement = document.CreateElement("span");
			XmlText link_text = document.CreateTextNode(text);
			linkElement.SetAttribute("class", "referenceNoLink");
			linkElement.AppendChild(link_text);
			return linkElement;
		}

	}

	internal enum ReferenceLinkType
	{
		None,
		Local,
		Index
	}

}