// Copyright (c) Microsoft Corporation.  All rights reserved.
//
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Xml;
using System.Xml.XPath;

// still have problems with spaces

namespace Microsoft.Ddue.Tools {

	public class SharedContentResolverComponent : BuildComponent {

		private XmlNamespaceManager nsManager = new XmlNamespaceManager(new NameTable());

		private XPathExpression sharedContentQuery;

		public SharedContentResolverComponent(XPathNavigator configuration)
			: base(configuration)
		{

			// get the source and target formats
			XPathNodeIterator content_nodes = configuration.Select("content");
			foreach (XPathNavigator content_node in content_nodes) {
				string file = content_node.GetAttribute("file", String.Empty);
				if ((file == null) || (file.Length == 0)) throw new ConfigurationErrorsException("When instantiating a shared content component, you must use the file attribute to specify a shared content file.");
				LoadContent(file);
			}

			WriteMessage(MessageLevel.Info, String.Format("Loaded {0} shared content items.", content.Count) );

			// set up the context
			// put in a default entry for ddue
			nsManager.AddNamespace("ddue", "http://ddue.schemas.microsoft.com/authoring/2003/6");
			// look for context nodes, which could override the default
			XPathNodeIterator contextNodes = configuration.Select("context");
			foreach (XPathNavigator contextNode in contextNodes)
			{
				string prefix = contextNode.GetAttribute("prefix", String.Empty);
				string uri = contextNode.GetAttribute("name", String.Empty);
				// add the namespace's prefix and uri. If the prefix exists in the manager, this replaces it.
				nsManager.AddNamespace(prefix, uri);
			}

			// initialize the xpath expression
			string xpathString = "//include | //includeAttribute | //ddue:token";
			// look for xpath nodes, which could override the default
			XPathNodeIterator xpathNodes = configuration.Select("xpath");
			foreach (XPathNavigator xpathNode in xpathNodes)
			{
				string expression = xpathNode.GetAttribute("expression", String.Empty);
				if (expression != string.Empty)
					xpathString = expression;
			}
			sharedContentQuery = XPathExpression.Compile(xpathString);

			// set context for the xpath expression
			sharedContentQuery.SetContext(nsManager);
		}

		private void LoadContent (string file) {

			WriteMessage(MessageLevel.Info, String.Format("Loading shared content file '{0}'.", file) );

			XmlReader reader = XmlReader.Create(file);
			try {
				reader.MoveToContent();
				while (!reader.EOF) {
	
					if ((reader.NodeType == XmlNodeType.Element) && (reader.Name == "item")) {

						string key = reader.GetAttribute("id");
						if (key != string.Empty)
						{
							string value = reader.ReadInnerXml();
							if (value == string.Empty)
								WriteMessage(MessageLevel.Warn, String.Format("Empty string in shared content item '{0}'.", key));
								//WriteMessage(MessageLevel.Warn, "loading", String.Format("Empty string in shared content item '{0}'.", key));
							
							// for duplicate item ids, the last one wins
							if (content.ContainsKey(key.ToLower()))
							{
								content.Remove(key.ToLower());
								WriteMessage(MessageLevel.Warn, string.Format("Duplicate shared content item '{0}'. Last one wins!", key));
								//WriteMessage(MessageLevel.Warn, "loading", string.Format("Duplicate shared content item '{0}'. Last one wins!", key));
							}
							// add the item
							content.Add(key.ToLower(), value);
						}
					} else {
						reader.Read();
					}

				}
			} finally {
				reader.Close();
			}

		}

		private Dictionary<string,string> content = new Dictionary<string,string>();

		private string currentKey = string.Empty;

		public override void Apply (XmlDocument document, string key) {
			currentKey = key;
			ResolveContent(document, document.CreateNavigator());
		}

		private void ResolveContent(XmlDocument document, XPathNavigator start)
		{
			// find all content inclusion instructions
			XPathNavigator includeNode = start.SelectSingleNode(sharedContentQuery);
			while (includeNode != null)
			{
				try
				{
					string nodeName = includeNode.LocalName;
					// get the key
					string key = includeNode.GetAttribute("item", String.Empty);
					string value;
					if (key == string.Empty)
					{
						WriteMessage(MessageLevel.Warn, string.Format("The '{0}' node has no item identifier. Topic id: '{1}'", nodeName, currentKey));
						//WriteMessage(MessageLevel.Warn, nodeName, string.Format("The '{0}' node has no item identifier. Topic id: '{1}'", nodeName, currentKey));
						includeNode.DeleteSelf();
					}
					else if (!content.TryGetValue(key.ToLower().Trim(), out value))
					{
						WriteMessage(MessageLevel.Warn, string.Format("The '{0}' node identifier '{1}' not found. Topic id: '{2}'", nodeName, key, currentKey));
						//WriteMessage(MessageLevel.Warn, nodeName, string.Format("The '{0}' node identifier '{1}' not found. Topic id: '{2}'", nodeName, key, currentKey));
						includeNode.DeleteSelf();
					}
					else if (value == string.Empty)
					{
						includeNode.DeleteSelf();
					}
					else
					{
						// got a key, got a value

						// extract any parameters
						List<string> parameters = new List<string>();
						XPathNodeIterator parameter_nodes = includeNode.Select("parameter");
						foreach (XPathNavigator parameter_node in parameter_nodes)
						{
							string parameter = parameter_node.InnerXml;
							parameters.Add(parameter);
						}

						// insert the parameter values into the shared content item
						string valueWithParams = string.Empty;
						try
						{
							valueWithParams = String.Format(value, parameters.ToArray());
						}
						catch (FormatException fe)
						{
							WriteMessage(MessageLevel.Error, string.Format("The '{0}' node with id '{1}' could not be formatted with {2} parameters.\n{3}", nodeName, key, parameters.Count, fe.Message));
						}

						// store the content in a document fragment
						XmlDocumentFragment fragment = document.CreateDocumentFragment();
						fragment.InnerXml = valueWithParams;

						// resolve any shared content in the fragment
						ResolveContent(document, fragment.CreateNavigator());
						string resolvedContent = fragment.InnerXml;

						if (resolvedContent.Trim() == string.Empty)
						{
							includeNode.DeleteSelf();
						}
						else
						{

							// replace the includeNode with the contents of the fragment
							switch (nodeName)
							{
								case "includeAttribute":
									string name = includeNode.GetAttribute("name", String.Empty);
									if (name != string.Empty)
									{
										XPathNavigator parent = includeNode.CreateNavigator();
										parent.MoveToParent();
										parent.CreateAttribute(String.Empty, name, String.Empty, resolvedContent);
									}
									else
									{
										WriteMessage(MessageLevel.Warn, string.Format("The 'includeAttribute' node with id '{0}' is missing @name to specify the attribute to set.", key));
										//WriteMessage(MessageLevel.Warn, "includeAttribute", string.Format("The 'includeAttribute' node with id '{0}' is missing @name to specify the attribute to set.", key));
									}
									includeNode.DeleteSelf();
									break;
								default:
									includeNode.ReplaceSelf(resolvedContent);
									break;
							}
						}
					}
					includeNode = start.SelectSingleNode(sharedContentQuery);
				}
				catch (Exception e)
				{
					WriteMessage(MessageLevel.Error, string.Format("Exception: '{0}'", e.Message));
				}
			}
		}

	}

}